#include "stdafx.h"
#include "PathMap.h"

const CPoint CPathMap::s_Dirs[9] = 
{
	CPoint(  1,  0 ),   // 0
	CPoint(  1,  1 ),   // 1
	CPoint(  0,  1 ),   // 2
	CPoint( -1,  1 ),   // 3
	CPoint( -1,  0 ),   // 4
	CPoint( -1, -1 ),   // 5
	CPoint(  0, -1 ),   // 6
	CPoint(  1, -1 ),   // 7
	CPoint(  0,  0 )    // 8
};

const char CPathMap::s_Dirs_English[9][4] = 
{
    "E", "SE", "S", "SW", "W", "NW", "N", "NE", "N/A"
};

CPathMap::CPathMap()
{
}

CPathMap::~CPathMap()
{

}

#pragma region The_cache_stuff
void CPathMap::CBestUnvisitedCache::Clear()
{
	for( int i=0; i<s_Size; i++ )
	{
		m_data[i].cost = u32( -1 );   // Maximum cost
		m_data[i].ind = size_t( -1 ); // Invalid index
	}
    m_BestOutOfCache = u32( -1 );   // Maximum cost
}

bool CPathMap::CBestUnvisitedCache::Push( size_t ind, u32 uProposedCost )
{
    // We don't add the item if it's worse then the best uncached value,
    // however we sometimes do add the items with the same cost.
    if( uProposedCost > m_BestOutOfCache )
        return false;

	int iFound = -1;     // index of the found value
	u32 uFoundCost = 0;  // cost of the found value
	for( int i=0; i<s_Size; i++ )
	{
		const u32 uThisCost = m_data[i].cost;
		if( m_data[i].ind == ind )	// the item is already on the cache..
		{
			if( uThisCost > uProposedCost )	//..and the new cost is better then the old one, so update
				m_data[i].cost = uProposedCost;
			return true;
		}
		if( uThisCost <= uProposedCost )
			continue;	// Currently stored item is better then the new one

		if( uThisCost <= uFoundCost ) // Previously we found worse value to replace
			continue;  // Remember, we're searching for the worst element to replace
		uFoundCost = uThisCost;
		iFound = i;
	}

	if( iFound < 0 )
    {
        // All the items on the cache are better then the proposed one
        if( uProposedCost < m_BestOutOfCache )
            m_BestOutOfCache = uProposedCost;
        return false;
    }
    // Update the m_BestOutOfCache if we're going to throw out a better value
    const u32 uCostToThrowOut = m_data[iFound].cost;
    if( uCostToThrowOut < m_BestOutOfCache )
        m_BestOutOfCache = uCostToThrowOut;

	m_data[iFound].cost = uProposedCost;
	m_data[iFound].ind = ind;
    return true;
}

bool CPathMap::CBestUnvisitedCache::Pop( size_t &ind )
{
	int iFound = -1;
	u32 uFoundVal = u32( -1 );
	for( int i=0; i<s_Size; i++ )
	{
		const u32 uCurr = m_data[i].cost;
		if( uCurr >= uFoundVal ) // A better deal is already found, or both are empty
			continue;
		iFound = i;
		uFoundVal = uCurr;
	}
	if( iFound < 0 )
		return false;

	ind = m_data[iFound].ind;
    // Cleanup the old cache entry
    m_data[iFound].cost = u32( -1 );   // Maximum cost
    m_data[iFound].ind = size_t( -1 ); // Invalid index
	return true;
}

size_t CPathMap::CBestUnvisitedCache::Refill( const CPathMap::SMapEntry* pData, size_t dataSize )
{
	m_BestOutOfCache = u32( -1 );   // Maximum cost
    size_t res = 0;
	for( size_t i=0; i<dataSize; i++ )
	{
		if( !pData[i].isHot() )
			continue;
		res++;
        const u32 costTotal = pData[i].costTotal;
		if( !Push( i, costTotal ) ) // not cached for some reason
        {
            if( costTotal < m_BestOutOfCache )
                m_BestOutOfCache = costTotal;
        }
	}
	return res;
}
#pragma endregion

void CPathMap::DijkstraProcess( size_t ind )
{
	SMapEntry& eThis = m_map[ind];
	if( eThis.isDone() || !eThis.isHot() ) __debugbreak();
	const CPoint ptThis = m_map.GetIndex( ind );
	
	for( int i=0; i<8; i++ )
	{
		const CPoint ptNeighbor = ptThis + s_Dirs[i];
		if( m_map.IsOutOfBounds( ptNeighbor ) )
			continue;

		const size_t indThat = m_map.GetIndex( ptNeighbor );
		SMapEntry& eThat = m_map[ indThat ];
		if( eThat.isDone() || eThat.cost == u16Max )	// Either already done on not passable
			continue;

		const u32 costProposed = GetStepCost( i, eThat.cost ) + eThis.costTotal;
		if( !eThat.isHot() || eThat.isDirty() )	// that pixel was not hot
		{
			eThat.setHot();
            eThat.clearDirty();
			eThat.costTotal = costProposed;
			eThat.directionForward = ( i + 4 ) & 7;
			m_BestUnvisited.Push( indThat, costProposed );
			continue;
		}
		// That pixel was hot already.
		const u32 costThatCurrent = eThat.costTotal;
		if( costThatCurrent <= costProposed )
			continue;	// It already has the better way out
		eThat.costTotal = costProposed;
		eThat.directionForward = ( i + 4 ) & 7;
		m_BestUnvisited.Push( indThat, costProposed );
	}
    eThis.flags = ( eThis.flags & eThis.FLAG_DEST ) | eThis.FLAG_DONE;
}

int CPathMap::UpdateWholeGraph()
{
    // Cleanup everything
    m_maxDanger = 0;
    for( size_t i=0; i<m_map.GetCount(); i++ )
    {
        SMapEntry& e = m_map[i];
        e.costTotal = u32Max;
        e.directionForward = 8;
        e.costTotal = 0;
        if( !e.isDestination() )
            e.flags = 0;
        else
            e.flags = e.FLAG_DEST | e.FLAG_HOT;
        if( e.cost > m_maxDanger )
            m_maxDanger = e.cost;
    }

    // Go on
    m_BestUnvisited.Clear();
    int iProcessed = 0;
    int iOverruns = 0;
    while( 1 )
    {
        size_t indNext;
        if( !m_BestUnvisited.Pop( indNext ) )
        {
            if( !m_BestUnvisited.Refill( m_map.GetData(), m_map.GetCount() ) )
                break;
            iOverruns++;
            if( !m_BestUnvisited.Pop( indNext ) )
                __debugbreak();
        }
        DijkstraProcess( indNext );
        iProcessed++;
    }
    MyTrace( "Pixels processed: %i; Cache refills: %i\n", iProcessed, iOverruns );
    m_bUpdatesNeeded = false;
    return iProcessed;
}

// Keep in mind that the values are added across the whole path,
// so don't return big values like 0x1000000..
inline u32 CPathMap::GetStepCost( int dir, u16 destCost ) const
{
    const u32 costMoveBase = 128;
    const u32 costMoveDiagonal = u32( double(costMoveBase) * sqrt(2.0) );
    // Below is the same code without double arithmetic
    //    const u32 costMove = u32(128) + u32( dir & 1 ) * 53;    // 53 = 128 * ( sqrt(2) - 1 )
    const u32 costMove = costMoveBase + u32( dir & 1 ) * ( costMoveDiagonal-costMoveBase );

    const u32 costArrive = u32(destCost);
    return costMove + costArrive * 4;
}

void CPathMap::updateMarkDirty( const CPoint& pt, bool bNowBlocked )
{
    m_map[pt].flags |= SMapEntry::FLAG_DIRTY;
    if( !bNowBlocked )
        return;

    // If the passability is changed,
    // we immediately replace the direction flags of this area and all neighbors to "none",
    // to avoid units going through the nearly built wall.
	// They better stand still for that several frames path map takes to rebuild.
    for( int i=0; i<9; i++ )
    {
        const CPoint ptThat = pt + s_Dirs[i];
        if( !m_map.IsOutOfBounds( ptThat ) )
            m_map[ptThat].directionForward = 8;
    }
}

int CPathMap::RecalcIncremental( int iMaxPixels, CRect& rcUpdatedArea )
{
    rcUpdatedArea = CRect(0,0,0,0);
    if( !m_bUpdatesNeeded )
        return 0;

    for( int i=0; i<iMaxPixels; i++ )
    {
        size_t indNext;
        if( !m_BestUnvisited.Pop( indNext ) )
        {
            if( !m_BestUnvisited.Refill( m_map.GetData(), m_map.GetCount() ) )
            {
                m_bUpdatesNeeded = false;
                return i;
            }

            if( !m_BestUnvisited.Pop( indNext ) )
                __debugbreak();
        }
        DijkstraProcess( indNext );

        // Calculate the updated area
        if( rcUpdatedArea.right == 0 )
        {
            const CPoint ptUpdatedTopLeft = m_map.GetIndex( indNext ) - CPoint( 1, 1 );
            rcUpdatedArea = CRect( ptUpdatedTopLeft, CSize( 3, 3 ) );
        }
        else
        {
            const CPoint ind = m_map.GetIndex( indNext );
            rcUpdatedArea.left = std::min( rcUpdatedArea.left, ind.x-1 );
            rcUpdatedArea.top = std::min( rcUpdatedArea.top, ind.y-1 );
            rcUpdatedArea.right = std::max( rcUpdatedArea.right, ind.x+2 );
            rcUpdatedArea.bottom = std::max( rcUpdatedArea.bottom, ind.y+2 );
        }
    }
    CRect rc2;
    rc2.IntersectRect(rcUpdatedArea, CRect( CPoint(0), m_map.GetSize() ) );
    rcUpdatedArea = rc2;
    return iMaxPixels;
}

int CPathMap::updateStart()
{
    // Pass #1
    // Determine the best (=closest to the destination) invalidated cell
    u32 bestDirtyArea = u32Max;
    for( size_t i=0; i<m_map.GetCount(); i++ )
    {
        SMapEntry& eThis = m_map[i];
        if( eThis.isDone() )
        {
            if( !eThis.isDirty() )
                continue;   // not dirty
        }
        else    // Not done, so mark as dirty
            eThis.setDirty();

        const u32 costTotal = eThis.costTotal;
        if( bestDirtyArea < costTotal )
            continue;
        bestDirtyArea = costTotal;
    }

    // Pass #2
    // setHot() all the cells that are neighbors to the cells who's costTotal >= bestDirtyArea
    // This is the most expensive 'for' here, that's why its body only sets the flag
    CPoint pThis;
    for( pThis.y = 0; pThis.y < m_map.GetSize().cy; pThis.y++ )
    {
        for( pThis.x = 0; pThis.x < m_map.GetSize().cx; pThis.x++ )
        {
            SMapEntry& eThis = m_map[pThis];
            const u32 costTotal = eThis.costTotal;
            if( costTotal < bestDirtyArea || costTotal == u32Max )
                continue;
            eThis.setHot();
            for( int i=0; i<8; i++ )
            {
                const CPoint pThat = pThis + s_Dirs[i];
                if( m_map.IsOutOfBounds( pThat ) )
                    continue;
                m_map[pThat].setHot();
            }
        }
    }

    // Pass #3
    int res = 0;
    for( size_t i=0; i < m_map.GetCount(); i++ )
    {
        SMapEntry& eThis = m_map[i];
        u8 flagsToRemove = 0;
        do 
        {
            if( !eThis.isHot() )    // not hot
                continue;

            flagsToRemove = SMapEntry::FLAG_DONE;

            if( u16Max == eThis.cost )  // Not passable
            {
                eThis.costTotal = u32Max;
                flagsToRemove |= SMapEntry::FLAG_HOT;
                continue;
            }
            res++;

            if( eThis.isDirty() || eThis.costTotal > bestDirtyArea )   // Was dirty
            {
                eThis.costTotal = u32Max;
                flagsToRemove |= SMapEntry::FLAG_HOT;
                flagsToRemove |= SMapEntry::FLAG_DIRTY;
                continue;
            }
            
        }
        while (0);

        eThis.flags &= ( ~flagsToRemove );
    }
    m_bUpdatesNeeded = true;
    m_BestUnvisited.Clear();

    return res;
}