#pragma once
#include "../Utils/Array2D.hpp"

#ifdef _WIN32
	typedef unsigned __int16 u16;
	typedef unsigned __int8 u8;
	typedef unsigned __int32 u32;
#endif

// Directions:
//  5  6  7
//  4  8  0
//  3  2  1
class CPathMap
{
public:
	CPathMap();
	~CPathMap();

    static const CPoint s_Dirs[9];
    static const char s_Dirs_English[9][4];

    // Mark the specified pixel as dirty.
    // Set bNowBlocked to immediately clean the previously-calculated directions of all neighbors cell
    void updateMarkDirty( const CPoint& pt, bool bNowBlocked );

protected:
    int updateStart();

    int RecalcIncremental( int iMaxPixels, CRect& rcUpdatedArea );

    // Dijkstra's algorithm for the whole graph
    // Returns the total count of processed pixels
    int UpdateWholeGraph();

	struct SMapEntry
	{
		// Total cost to the target, including the whole path,
        // or u32Max for non-passable cells
        u32  costTotal;

        // Cost to step onto this : 0 = free, -1 = non-passable, >0 = dangerous
        u16  cost;
		
        // Where to move from this point: from 0 to 7, 0 = East, 2 = South, 3 = South-West
        u8   directionForward;

        // Flags, see below
        u8   flags;

        static const u8 FLAG_HOT = 1;
        static const u8 FLAG_DONE = 2;
        static const u8 FLAG_DEST = 16;

        // This bit is set on the pixels that was just updated
        // This bit is cleared during the DoUpdateArea() call
        static const u8 FLAG_DIRTY = 64;

        inline bool isHot() const { return flags & FLAG_HOT; }
        inline void setHot() { flags |= FLAG_HOT; }
        inline void clearHot() { flags &= (~FLAG_HOT);  }

        inline bool isDone() const { return 0 != ( flags & FLAG_DONE ); }
        inline void setDone() { flags |= FLAG_DONE; }
        inline void clearDone() { flags &= (~FLAG_DONE); }

        inline bool isDestination() const { return 0 != ( flags & FLAG_DEST ); }
        inline void setDestination() { flags |= FLAG_DEST; }

        inline bool isDirty() const { return 0 != ( flags & FLAG_DIRTY ); }
        inline void setDirty() { flags |= FLAG_DIRTY; }
        inline void clearDirty() { flags &= (~FLAG_DIRTY); }
	};

	// The map itself
    CArray2D<SMapEntry> m_map;

    // max. danger found on the pasable cells
    u16 m_maxDanger;
    // Is set to true when we need to update something
    bool m_bUpdatesNeeded;

	// This class implements a cache for best unvisited pixels.
    // If it's empty, just call Refill passing the data stored in m_map.
    class CBestUnvisitedCache
	{
		static const int s_Size = 8;
        // the cache itself
		struct
		{
			size_t ind;
			u32 cost;
		}
		m_data[s_Size];
        // Cost of the best item that is not fitted in the cache
        u32 m_BestOutOfCache;
	public:
		// Cleanup the cache
        void Clear();

		// Add a value if it's better (=less) then any existing one,
        // replacing the worst item found on the cache
        // return false if not fit
		bool Push( size_t ind, u32 cost );

		// Pop the best (=least) value from the cache;
		// return false if the cache is empty
		bool Pop( size_t &ind );

		// Re-fill the cache from the data map.
		// Only the pixels where isHot=true counts.
		// Returns the total count of hot pixels.
		size_t Refill( const SMapEntry* pData, size_t dataSize );
	}
	m_BestUnvisited;

	// Process the single pixel
	// The pixel must have isDone=false ( will be set to true afterwards ),
    // and isHot=true ( will be set to false afterwards )
	void DijkstraProcess( size_t ind );

    // Calculate the cost for a single step.
    // dir is from 0 to 7, destCost is from 0 to m_maxDanger
    inline u32 GetStepCost( int dir, u16 destCost ) const;

	static const u32 u32Max = (u32)(-1);
	static const u32 u16Max = (u16)(-1);
};