#include "stdafx.h"
#include "PathMapWin.h"
#include "../Utils/BitmapLoader.h"

HRESULT CPathMapWin::LoadBmp( LPCTSTR strFileName, bool loadDangerInfo )
{
    CArray2D<RGBTRIPLE> bm;
    HRESULT hr = LoadBitmapFile( bm, strFileName );
    if(FAILED(hr))
        return hr;

    if( loadDangerInfo )
    {
        hr = LoadBmpImpl( bm, &CPathMapWin::cellFromColor_Direct );
    }
    else
    {
        hr = LoadBmpImpl( bm, &CPathMapWin::cellFromColor_ToMaster );
        UpdateMapFromMaster();
    }
    if( FAILED( hr ) )
        printf( "Unable to read the input image\n" );
    UpdateWholeGraph();
    return S_OK;
}

HRESULT CPathMapWin::LoadBmp( UINT iResID )
{
    CArray2D<RGBTRIPLE> bm;
    HRESULT hr = LoadBitmapResource( bm, iResID );
    if(FAILED(hr))
        return hr;
    hr = LoadBmpImpl( bm, &CPathMapWin::cellFromColor_ToMaster );
    if( FAILED( hr ) )
        printf( "Unable to read the input image\n" );
    UpdateMapFromMaster();
    UpdateWholeGraph();
    return S_OK;
}

HRESULT CPathMapWin::LoadBmpImpl( const CArray2D<RGBTRIPLE> &bm, cellFromColorProc proc )
{
    const CSize sz(bm.GetSize());
    if( !m_map.SetSize( sz ) || !m_mapMasterCopy.SetSize( sz ) )
        return E_OUTOFMEMORY;
    m_map.ZeroMem();
    m_mapMasterCopy.ZeroMem();

    CPoint p;
    size_t dest=0;
    for( p.y=0; p.y<sz.cy; p.y++ )
    {
        for( p.x=0; p.x<sz.cx; p.x++ )
            (this->*proc)( p, bm[p] );
    }
    return S_OK;
}

void CPathMapWin::cellFromColor_Direct( const CPoint &cell, RGBTRIPLE c )
{
    static const BYTE wThreshold = 255 - s_BlackColorThreshold;
    SMapEntry res;
    ZeroMemory( &res, sizeof( res ) );
    do 
    {
        if( c.rgbtRed < s_BlackColorThreshold && c.rgbtGreen < s_BlackColorThreshold )
        {
            if( c.rgbtBlue < s_BlackColorThreshold )	// Black
            {
                res.cost = -1;
                break;
            }
            if( c.rgbtBlue > wThreshold )	// Blue
            {
                res.setDestination();
                break;
            }
        }
        // Any other
        if( c.rgbtRed > wThreshold && c.rgbtGreen > wThreshold && c.rgbtBlue > wThreshold )	// White :-)
            break;

        int iRedAmount = int(c.rgbtRed) * 2 - c.rgbtBlue - c.rgbtGreen;
        if( iRedAmount <=0 )
            res.cost = 0;
        else
            res.cost = iRedAmount>>1;
    }
    while(0);

    m_map[cell] = res;
}

void CPathMapWin::cellFromColor_ToMaster( const CPoint &cell, RGBTRIPLE c )
{
    static const BYTE wThreshold = 255 - s_BlackColorThreshold;
    if( c.rgbtRed < s_BlackColorThreshold && c.rgbtGreen < s_BlackColorThreshold )
    {
        if( c.rgbtBlue < s_BlackColorThreshold )	// Black
        {
            m_mapMasterCopy[cell] = CELL_IMPASABLE;
            return;
        }
        if( c.rgbtBlue > wThreshold )	// Blue
        {
            m_mapMasterCopy[cell] = CELL_DESTINATION;
            return;
        }
    }
    m_mapMasterCopy[cell] = CELL_PASABLE;
}

void CPathMapWin::UpdateMapFromMaster()
{
    // Copy the map
    m_TurretsCount.SetSize(m_map.GetSize());
    m_TurretsCount.ZeroMem();
    m_map.ZeroMem();
    for( size_t i=0; i < m_map.GetCount(); i++ )
    {
        const E_BASE_MAP_ENTRY b = m_mapMasterCopy[i];
        SMapEntry& me = m_map[i];
        switch(b)
        {
        case CELL_IMPASABLE:
            me.cost = u16Max;
            break;
        case CELL_DESTINATION:
            me.setDestination();
            break;
        case CELL_PASABLE:
            break;
        default:
            __debugbreak();
        }
    }

    // Cleanup the turrets
	m_Turrets.SetCount(0);

    // Update the whole graph
    UpdateWholeGraph();
}

bool CPathMapWin::isTurretBody( CPoint pt, size_t* pTurrentIdx ) const
{
    for( size_t i = 0; i<m_Turrets.GetCount(); i++ )
    {
        const TurretInfo& t = m_Turrets[i];
        const CRect body( t.Position, CSize(2,2) );
        if( body.PtInRect( pt ) )
        {
            if( NULL != pTurrentIdx )
                *pTurrentIdx = i;
            return true;
        }
    }
    if( NULL != pTurrentIdx )
        *pTurrentIdx = -1;
    return false;
}

void CPathMapWin::AddTurret( const TurretInfo &t )
{
    m_Turrets.Add( t );
    processTurret( t, &CPathMapWin::turret_Add );
    updateStart();
}

void CPathMapWin::RemoveTurret( size_t indTurret )
{
    TurretInfo ti = m_Turrets[indTurret];
    m_Turrets.RemoveAt(indTurret);

    processTurret( ti, &CPathMapWin::turret_Remove );

    for( size_t i=0; i<m_mapMasterCopy.GetCount(); i++ )
    {
        if( CELL_PASABLE != m_mapMasterCopy[i] )
            continue;
        if( m_TurretsCount[i] != 0 )
            continue;
        m_map[i].cost = 0;
    }

    for( int i=0; i<m_Turrets.GetCount(); i++ )
        processTurret( m_Turrets[i], &CPathMapWin::turret_UpdateCoverage );

    updateStart();
}

void CPathMapWin::processTurret( const TurretInfo& ti, turretUpdateProc proc )
{
    const CRect wholeMap( CPoint(0), m_map.GetSize() );
    const CRect body( ti.Position, CSize( 2,2 ) );

    CRect damage = body;
    damage.InflateRect( ti.Radius, ti.Radius );
    const long _4R2 = ti.Radius * ti.Radius * 4;

    CPoint p;
    for( p.y = damage.top; p.y < damage.bottom; p.y++ )
        for( p.x=damage.left; p.x < damage.right; p.x++ ) 
    {
        if( !wholeMap.PtInRect( p ) )
            continue;   // out of map
        const bool isBody = body.PtInRect ( p );
        const long _4currRadius2 = 
            Sqr( ti.Position.x * 2 + 1 - p.x * 2 ) +
            Sqr( ti.Position.y * 2 + 1 - p.y * 2 );
        if( _4currRadius2 > _4R2 )
            continue;
        if( CELL_PASABLE != m_mapMasterCopy [p] )
            continue;
        (this->*proc)( p, isBody, ti.Damage );
    }
}

void CPathMapWin::turret_Add( const CPoint &cell, bool isBody, BYTE damage )
{
    SMapEntry& e = m_map[cell];
    if( e.cost == u16Max )
        return;
    if( isBody )
    {
        e.cost = u16Max;
        updateMarkDirty( cell, true );
        m_TurretsCount[cell]++;
        return;
    }
    e.cost += damage;
    updateMarkDirty( cell, false );
}

void CPathMapWin::turret_Remove( const CPoint &cell, bool isBody, BYTE damage )
{
    SMapEntry& e = m_map[cell];
    
    if( isBody )
    {
        m_TurretsCount[cell]--;
        if( 0 == m_TurretsCount[cell] )
        {
            m_map[cell].cost = 0;
            m_map[cell].costTotal = u32Max - 1;
        }
        updateMarkDirty( cell, false );
    }
    else
        updateMarkDirty( cell, false );
}

void CPathMapWin::turret_UpdateCoverage( const CPoint &cell, bool isBody, BYTE damage )
{
    m_map[cell].cost += damage;
}

CString CPathMapWin::GetCellInfo( const CPoint& pt ) const
{
    if( m_map.IsOutOfBounds(pt) )
        return L"Out of bounds";

    const SMapEntry& e = m_map[pt];

    CString res;
    res.Format(L"X: %i, Y: %i, Index: %i; ",
        int( pt.x ), int( pt.y ), int( m_map.GetIndex( pt ) ) );

    if( e.cost == u16Max )
        res += L"Cost: max; ";
    else
        res.AppendFormat( L"Cost: %i; ", int(e.cost) );

    if( e.costTotal == u32Max )
        res += L"Total cost: max; ";
    else
        res.AppendFormat( L"Total cost: %i; ", int( e.costTotal ) );

    res.AppendFormat(L"Direction: %S; ", s_Dirs_English[ e.directionForward ] );

    res += L"Flags: ";
    const struct FlagIn4
    {
        LPCTSTR strName; BYTE bFlag;
    }
    arrFlags[4] = 
    {
        L"hot", SMapEntry::FLAG_HOT,
        L"done", SMapEntry::FLAG_DONE,
        L"dest", SMapEntry::FLAG_DEST,
        L"dirty", SMapEntry::FLAG_DIRTY,
    };
    for( int i=0; i<4; i++ )
    {
        FlagIn4 fi = arrFlags[i];
        res += fi.strName;
        res += L"=";
        res += ( e.flags & fi.bFlag ) ? '1' : '0';
        res += L"; ";
    }
    return res;
}

void CPathMapWin::UpdateWholeGraph()
{
    const QWORD qwStart = __rdtsc();
    const int iPixelsProcessed = __super::UpdateWholeGraph();
    const QWORD qwEnd = __rdtsc();

    MyTrace( "\tUpdateWholeGraph called.\n" );
    MyTrace( "Cells affected: %i; CPU ticks: %i; time @ 600 MHZ: %g ms\n",
        iPixelsProcessed,
        int( qwEnd - qwStart ),
        double( qwEnd - qwStart ) * 1000.0 / 600E6 );
}

int CPathMapWin::RecalcIncremental( int iMaxPixels, CRect& rcUpdatedArea )
{
    const QWORD qwStart = __rdtsc();
    const int res = __super::RecalcIncremental( iMaxPixels, rcUpdatedArea );
    const QWORD qwEnd = __rdtsc();

    MyTrace( "\tRecalcIncremental called.\n" );
    MyTrace( "Cells affected: %i; CPU ticks: %i; time @ 600 MHZ: %g ms\n",
        res,
        int(qwEnd-qwStart),
        double(qwEnd-qwStart) * 1000.0 / 600E6 );
    return res;
}

void CPathMapWin::updateStart()
{
    const QWORD qwStart = __rdtsc();
    const int res = __super::updateStart();
    const QWORD qwEnd = __rdtsc();

    MyTrace( "\tupdateStart called.\n" );
    MyTrace( "Cells affected: %i; CPU ticks: %i; time @ 600 MHZ: %g ms\n",
        res,
        int(qwEnd-qwStart),
        double(qwEnd-qwStart) * 1000.0 / 600E6 );
}