#pragma once
#include "MouseProcessor.h"

template<class T>
class CGridViewWnd:
    public CWindowImpl<T, CWindow, CControlWinTraits>,
    public CMouseProcessor
{
public:
    CGridViewWnd(){}
    ~CGridViewWnd(){}

    // Create a grid view of the specified size
    HRESULT Initialize( CSize sz );

    static const CSize s_Cell;

    static const DWORD dwClassStyle = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;

    BEGIN_MSG_MAP( CGridViewWnd<T> )
        // General
        MESSAGE_HANDLER( WM_SIZE, OnSize );
        MESSAGE_HANDLER( WM_PAINT, OnPaint );

        CHAIN_MSG_MAP(CMouseProcessor);
    END_MSG_MAP()
protected:
    void RecalcLayout();
private:
    LRESULT OnSize( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled )
    {
        RecalcLayout();
        return 0;
    }
    LRESULT OnPaint( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled );

    CSize m_cellsSize;
    CPoint m_clientTopLeft;
    CRect m_clientRect;

    // Returns the pixel size of the image
    inline CSize pxGetClientSize() const
    {
        //TODO: scale here
        return CSize( m_cellsSize.cx * s_Cell.cx,
                        m_cellsSize.cy * s_Cell.cy );
    }

    // Returns the cells that are within the rectangle in window client coordinates
    inline CRect cellsFromClient( CRect pxRect) const
    {
        pxRect.OffsetRect(-m_clientTopLeft);
        //TODO: scale here with CRect::MulDiv
        CRect res;
        res.left = pxRect.left / s_Cell.cx;
        res.top = pxRect.top / s_Cell.cy;
        res.right = ( pxRect.right + s_Cell.cx - 1 ) / s_Cell.cx;
        res.bottom = ( pxRect.bottom + s_Cell.cy - 1 ) / s_Cell.cy;
        const CRect cellsTotal( CPoint(0), m_cellsSize );
        return IntersectRect( res, cellsTotal );
    }

    inline CRect pxGetCellRect( CPoint cell, CSize sz = CSize(0,0) )
    {
        CRect res( cell, sz + CSize(1,1) );
        res.left *= s_Cell.cx;
        res.right *= s_Cell.cx;
        res.top *= s_Cell.cy;
        res.bottom *= s_Cell.cy;
        //TODO: scale here with CRect::MulDiv
        res.OffsetRect(m_clientTopLeft);
        return res;
    }

    // The ::IntersectRect fails when source==destination,
    // so CRect::IntersectRect fails when this==Rect1 or this==Rect2 :-(
    inline CRect IntersectRect(const CRect& r1, const CRect& r2) const
    {
        CRect res;
        res.IntersectRect(r1, r2);
        return res;
    }

    // The HitTest implementation for CMouseProcessor
	virtual int HitTest( CPoint pt )
    {
        if( m_cellsSize.cx <=0 || m_cellsSize.cy <=0 )
            return -1;
        const CRect clientImage( m_clientTopLeft, pxGetClientSize() );
        if(!clientImage.PtInRect(pt))
            return -1;
        pt -= m_clientTopLeft;
        int x = pt.x / s_Cell.cx;
        int y = pt.y / s_Cell.cy;
        return x + y * m_cellsSize.cx;
    }
protected:
    inline void RepaintCell( CPoint ptCell )
    {
        const CRect rcCell = pxGetCellRect(ptCell);
        RedrawWindow(rcCell);
    }
    inline void InvalidateCells( const CRect &rxCells )
    {
        const CRect rcCell = pxGetCellRect(rxCells.TopLeft(), rxCells.Size());
        InvalidateRect(rcCell);
    }
};

template<class T>
const CSize CGridViewWnd<T>::s_Cell( 12, 12 );

template<class T>
void CGridViewWnd<T>::RecalcLayout()
{
    GetClientRect( &m_clientRect );
    CSize size = pxGetClientSize();
    size.cx /= 2;
    size.cy /= 2;
    m_clientTopLeft = m_clientRect.CenterPoint() - size;
}

template<class T>
LRESULT CGridViewWnd<T>::OnPaint( UINT nMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled )
{
    if( m_cellsSize.cx <=0 || m_cellsSize.cy <=0 )
    {
        bHandled = FALSE;
        return 0;
    }

    PAINTSTRUCT ps;
    HDC dc = BeginPaint(&ps);

    // Calculate dirty region in client coordinates
    const CRect clientImage( m_clientTopLeft, pxGetClientSize() );
    CRect clientDirty = IntersectRect(m_clientRect, clientImage);
    clientDirty = IntersectRect( clientDirty, ps.rcPaint );
    if( clientDirty.IsRectEmpty() )
    {
        EndPaint(&ps);
        return 0;
    }

    const CRect cellsDirty = cellsFromClient( clientDirty );
    CPoint c;
    for( c.y = cellsDirty.top; c.y < cellsDirty.bottom; c.y++ )
    {
        for( c.x = cellsDirty.left; c.x < cellsDirty.right; c.x++ )
        {
            const CRect rc = pxGetCellRect(c);
            static_cast<T*>(this)->PaintCell( dc, c, rc );
        }
    }

    EndPaint(&ps);
    return 0;
}

template<class T>
HRESULT CGridViewWnd<T>::Initialize( CSize sz )
{
    if( sz.cx <=0 || sz.cy <= 0 )
        return E_INVALIDARG;
    m_cellsSize = sz;
    return S_OK;
}